from enum import IntEnum
from typing import Callable, Dict, List, Iterable

from spherov2.commands.animatronic import R2LegActions, Animatronic
from spherov2.commands.api_and_shell import ApiAndShell
from spherov2.commands.core import IntervalOptions, Core
from spherov2.commands.io import AudioPlaybackModes, IO
from spherov2.commands.power import Power
from spherov2.commands.sensor import CollisionDetectionMethods, Sensor, SensitivityBasedCollisionDetectionMethods, \
    SensitivityLevels
from spherov2.commands.sphero import CollisionDetectionMethods as SpheroCollisionDetectionMethods, Sphero
from spherov2.controls import RawMotorModes
from spherov2.controls.v2 import Processors
from spherov2.toy import Toy
from spherov2.toy.bb9e import BB9E
from spherov2.toy.bolt import BOLT
from spherov2.toy.mini import Mini
from spherov2.toy.r2d2 import R2D2
from spherov2.toy.r2q5 import R2Q5
from spherov2.toy.rvr import RVR


class ToyUtil:
    @staticmethod
    def sleep(toy: Toy, not_supported_handler: Callable[[], None] = None):
        if toy.implements(Core.sleep):
            toy.sleep(IntervalOptions.NONE, 0, 0)
        elif toy.implements(Power.sleep):
            toy.sleep()
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def ping(toy: Toy, not_supported_handler: Callable[[], None] = None):
        if toy.implements(Core.ping):
            toy.ping()
        elif toy.implements(ApiAndShell.ping):
            toy.ping(None)
        elif toy.implements(ApiAndShell.ping, True):
            toy.ping(None, Processors.PRIMARY)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def roll_start(toy: Toy, heading: int, speed: int, not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'drive_control'):
            toy.drive_control.roll_start(heading, speed)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def roll_stop(toy: Toy, heading: int, is_reverse: bool, not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'drive_control'):
            if is_reverse:
                heading = (heading + 180) % 360
            toy.drive_control.roll_stop(heading)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def perform_leg_action(toy: Toy, leg_action: R2LegActions, not_supported_handler: Callable[[], None] = None):
        if toy.implements(Animatronic.perform_leg_action):
            toy.perform_leg_action(leg_action)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def set_stabilization(toy: Toy, stabilize, not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'drive_control'):
            toy.drive_control.set_stabilization(stabilize)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def set_raw_motor(toy: Toy, left_mode: RawMotorModes, left_speed: int, right_mode: RawMotorModes, right_speed: int,
                      not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'drive_control'):
            toy.drive_control.set_raw_motors(left_mode, left_speed, right_mode, right_speed)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def reset_heading(toy: Toy, not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'drive_control'):
            toy.drive_control.reset_heading()
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def play_animation(toy: Toy, animation: IntEnum, wait: bool = False,
                       not_supported_handler: Callable[[], None] = None):
        if toy.implements(Animatronic.play_animation):
            toy.play_animation(animation, wait)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def set_head_position(toy: Toy, head_position: float, not_supported_handler: Callable[[], None] = None):
        if toy.implements(Animatronic.set_head_position):
            toy.set_head_position(head_position)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def set_main_led(toy: Toy, r: int, g: int, b: int, is_user_color: bool,
                     not_supported_handler: Callable[[], None] = None):
        def _fallback():
            if isinstance(toy, (R2D2, R2Q5)):
                mapping = {
                    toy.LEDs.BACK_RED: r,
                    toy.LEDs.BACK_GREEN: g,
                    toy.LEDs.BACK_BLUE: b,
                    toy.LEDs.FRONT_RED: r,
                    toy.LEDs.FRONT_GREEN: g,
                    toy.LEDs.FRONT_BLUE: b
                }
            elif isinstance(toy, BB9E):
                mapping = {
                    toy.LEDs.BODY_RED: r,
                    toy.LEDs.BODY_GREEN: g,
                    toy.LEDs.BODY_BLUE: b
                }
            elif isinstance(toy, Mini):
                mapping = {
                    toy.LEDs.BODY_RED: r,
                    toy.LEDs.BODY_GREEN: g,
                    toy.LEDs.BODY_BLUE: b,
                    toy.LEDs.USER_BODY_RED: r,
                    toy.LEDs.USER_BODY_GREEN: g,
                    toy.LEDs.USER_BODY_BLUE: b
                }
            elif isinstance(toy, RVR):
                mapping = {
                    toy.LEDs.RIGHT_HEADLIGHT_RED: r,
                    toy.LEDs.RIGHT_HEADLIGHT_GREEN: g,
                    toy.LEDs.RIGHT_HEADLIGHT_BLUE: b,
                    toy.LEDs.LEFT_HEADLIGHT_RED: r,
                    toy.LEDs.LEFT_HEADLIGHT_GREEN: g,
                    toy.LEDs.LEFT_HEADLIGHT_BLUE: b,
                    toy.LEDs.LEFT_STATUS_INDICATION_RED: r,
                    toy.LEDs.LEFT_STATUS_INDICATION_GREEN: g,
                    toy.LEDs.LEFT_STATUS_INDICATION_BLUE: b,
                    toy.LEDs.RIGHT_STATUS_INDICATION_RED: r,
                    toy.LEDs.RIGHT_STATUS_INDICATION_GREEN: g,
                    toy.LEDs.RIGHT_STATUS_INDICATION_BLUE: b,
                    toy.LEDs.BATTERY_DOOR_FRONT_RED: r,
                    toy.LEDs.BATTERY_DOOR_FRONT_GREEN: g,
                    toy.LEDs.BATTERY_DOOR_FRONT_BLUE: b,
                    toy.LEDs.BATTERY_DOOR_REAR_RED: r,
                    toy.LEDs.BATTERY_DOOR_REAR_GREEN: g,
                    toy.LEDs.BATTERY_DOOR_REAR_BLUE: b,
                    toy.LEDs.POWER_BUTTON_FRONT_RED: r,
                    toy.LEDs.POWER_BUTTON_FRONT_GREEN: g,
                    toy.LEDs.POWER_BUTTON_FRONT_BLUE: b,
                    toy.LEDs.POWER_BUTTON_REAR_RED: r,
                    toy.LEDs.POWER_BUTTON_REAR_GREEN: g,
                    toy.LEDs.POWER_BUTTON_REAR_BLUE: b,
                    toy.LEDs.LEFT_BRAKELIGHT_RED: r,
                    toy.LEDs.LEFT_BRAKELIGHT_GREEN: g,
                    toy.LEDs.LEFT_BRAKELIGHT_BLUE: b,
                    toy.LEDs.RIGHT_BRAKELIGHT_RED: r,
                    toy.LEDs.RIGHT_BRAKELIGHT_GREEN: g,
                    toy.LEDs.RIGHT_BRAKELIGHT_BLUE: b
                }
            else:
                mapping = None

            def __fallback():
                if toy.implements(Sphero.set_main_led):
                    toy.set_main_led(r, g, b)
                elif not_supported_handler:
                    not_supported_handler()

            ToyUtil.set_multiple_leds(toy, mapping, __fallback)

        ToyUtil.set_led_matrix_one_colour(toy, r, g, b, _fallback)

    @staticmethod
    def set_head_led(toy: Toy, brightness: int, not_supported_handler: Callable[[], None] = None):
        if isinstance(toy, BB9E):
            ToyUtil.set_multiple_leds(toy, {BB9E.LEDs.HEAD: brightness}, not_supported_handler)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def set_front_led(toy: Toy, r: int, g: int, b: int, not_supported_handler: Callable[[], None] = None):
        if isinstance(toy, RVR):
            mapping = {
                RVR.LEDs.RIGHT_HEADLIGHT_RED: r,
                RVR.LEDs.RIGHT_HEADLIGHT_GREEN: g,
                RVR.LEDs.RIGHT_HEADLIGHT_BLUE: b,
                RVR.LEDs.LEFT_HEADLIGHT_RED: r,
                RVR.LEDs.LEFT_HEADLIGHT_GREEN: g,
                RVR.LEDs.LEFT_HEADLIGHT_BLUE: b
            }
        elif isinstance(toy, (R2D2, R2Q5, BOLT)):
            mapping = {
                toy.LEDs.FRONT_RED: r,
                toy.LEDs.FRONT_GREEN: g,
                toy.LEDs.FRONT_BLUE: b
            }
        elif isinstance(toy, Mini):
            mapping = {
                toy.LEDs.BODY_RED: r,
                toy.LEDs.BODY_GREEN: g,
                toy.LEDs.BODY_BLUE: b
            }
        else:
            mapping = None
        ToyUtil.set_multiple_leds(toy, mapping, not_supported_handler)

    @staticmethod
    def set_back_led(toy: Toy, r: int, g: int, b: int, not_supported_handler: Callable[[], None] = None):
        if isinstance(toy, RVR):
            mapping = {
                RVR.LEDs.RIGHT_BRAKELIGHT_RED: r,
                RVR.LEDs.RIGHT_BRAKELIGHT_GREEN: g,
                RVR.LEDs.RIGHT_BRAKELIGHT_BLUE: b,
                RVR.LEDs.LEFT_BRAKELIGHT_RED: r,
                RVR.LEDs.LEFT_BRAKELIGHT_GREEN: g,
                RVR.LEDs.LEFT_BRAKELIGHT_BLUE: b
            }
        elif isinstance(toy, (R2D2, R2Q5, BOLT)):
            mapping = {
                toy.LEDs.BACK_RED: r,
                toy.LEDs.BACK_GREEN: g,
                toy.LEDs.BACK_BLUE: b
            }
        elif isinstance(toy, Mini):
            mapping = {
                toy.LEDs.USER_BODY_RED: r,
                toy.LEDs.USER_BODY_GREEN: g,
                toy.LEDs.USER_BODY_BLUE: b
            }
        else:
            mapping = None
        ToyUtil.set_multiple_leds(toy, mapping, not_supported_handler)

    @staticmethod
    def set_back_led_brightness(toy: Toy, brightness: int, not_supported_handler: Callable[[], None] = None):
        if isinstance(toy, (R2D2, R2Q5, BOLT)):
            mapping = {
                toy.LEDs.BACK_RED: 0,
                toy.LEDs.BACK_GREEN: 0,
                toy.LEDs.BACK_BLUE: brightness,
            }
        elif isinstance(toy, (BB9E, Mini)):
            mapping = {
                toy.LEDs.AIMING: brightness
            }
        elif isinstance(toy, RVR):
            mapping = {
                RVR.LEDs.RIGHT_BRAKELIGHT_RED: 0,
                RVR.LEDs.RIGHT_BRAKELIGHT_GREEN: 0,
                RVR.LEDs.RIGHT_BRAKELIGHT_BLUE: brightness,
                RVR.LEDs.LEFT_BRAKELIGHT_RED: 0,
                RVR.LEDs.LEFT_BRAKELIGHT_GREEN: 0,
                RVR.LEDs.LEFT_BRAKELIGHT_BLUE: brightness
            }
        else:
            mapping = None

        def _fallback():
            if toy.implements(Sphero.set_back_led_brightness):
                toy.set_back_led_brightness(brightness)
            elif not_supported_handler:
                not_supported_handler()

        ToyUtil.set_multiple_leds(toy, mapping, _fallback)

    @staticmethod
    def set_left_front_led(toy: Toy, r: int, g: int, b: int, not_supported_handler: Callable[[], None] = None):
        mapping = None
        if isinstance(toy, RVR):
            mapping = {
                RVR.LEDs.LEFT_HEADLIGHT_RED: r,
                RVR.LEDs.LEFT_HEADLIGHT_GREEN: g,
                RVR.LEDs.LEFT_HEADLIGHT_BLUE: b
            }
        ToyUtil.set_multiple_leds(toy, mapping, not_supported_handler)

    @staticmethod
    def set_right_front_led(toy: Toy, r: int, g: int, b: int, not_supported_handler: Callable[[], None] = None):
        mapping = None
        if isinstance(toy, RVR):
            mapping = {
                RVR.LEDs.RIGHT_HEADLIGHT_RED: r,
                RVR.LEDs.RIGHT_HEADLIGHT_GREEN: g,
                RVR.LEDs.RIGHT_HEADLIGHT_BLUE: b
            }
        ToyUtil.set_multiple_leds(toy, mapping, not_supported_handler)

    @staticmethod
    def set_battery_side_led(toy: Toy, r: int, g: int, b: int, not_supported_handler: Callable[[], None] = None):
        mapping = None
        if isinstance(toy, RVR):
            mapping = {
                RVR.LEDs.BATTERY_DOOR_FRONT_RED: r,
                RVR.LEDs.BATTERY_DOOR_FRONT_GREEN: g,
                RVR.LEDs.BATTERY_DOOR_FRONT_BLUE: b
            }
        ToyUtil.set_multiple_leds(toy, mapping, not_supported_handler)

    @staticmethod
    def set_power_side_led(toy: Toy, r: int, g: int, b: int, not_supported_handler: Callable[[], None] = None):
        mapping = None
        if isinstance(toy, RVR):
            mapping = {
                RVR.LEDs.POWER_BUTTON_FRONT_RED: r,
                RVR.LEDs.POWER_BUTTON_FRONT_GREEN: g,
                RVR.LEDs.POWER_BUTTON_FRONT_BLUE: b
            }
        ToyUtil.set_multiple_leds(toy, mapping, not_supported_handler)

    @staticmethod
    def set_holo_projector(toy: Toy, brightness: int, not_supported_handler: Callable[[], None] = None):
        if isinstance(toy, (R2D2, R2Q5)):
            mapping = {toy.LEDs.HOLO_PROJECTOR: brightness}
        else:
            mapping = None
        ToyUtil.set_multiple_leds(toy, mapping, not_supported_handler)

    @staticmethod
    def set_logic_display(toy: Toy, brightness: int, not_supported_handler: Callable[[], None] = None):
        if isinstance(toy, (R2D2, R2Q5)):
            mapping = {toy.LEDs.LOGIC_DISPLAYS: brightness}
        else:
            mapping = None
        ToyUtil.set_multiple_leds(toy, mapping, not_supported_handler)

    @staticmethod
    def set_multiple_leds(toy: Toy, mapping: Dict[IntEnum, int], not_supported_handler: Callable[[], None] = None):
        if mapping and hasattr(toy, 'multi_led_control'):
            toy.multi_led_control.set_leds(mapping)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def set_led_matrix_one_colour(toy: Toy, r: int, g: int, b: int, not_supported_handler: Callable[[], None] = None):
        if toy.implements(IO.set_compressed_frame_player_one_color):
            toy.set_compressed_frame_player_one_color(r, g, b)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def set_matrix_pixel(toy: Toy, x: int, y: int, r: int, g: int, b: int, is_user_color: bool,
                         not_supported_handler: Callable[[], None] = None):
        def _fallback():
            not_supported_handler()

        ToyUtil.set_led_matrix_pixel(toy, x, y, r, g, b, _fallback)

    @staticmethod
    def set_led_matrix_pixel(toy: Toy, x: int, y: int, r: int, g: int, b: int,
                             not_supported_handler: Callable[[], None] = None):
        if toy.implements(IO.set_compressed_frame_player_pixel):
            toy.set_compressed_frame_player_pixel(x, y, r, g, b)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def set_matrix_line(toy: Toy, x1: int, y1: int, x2: int, y2: int, r: int, g: int, b: int, is_user_color: bool,
                        not_supported_handler: Callable[[], None] = None):
        def _fallback():
            not_supported_handler()

        ToyUtil.set_led_matrix_line(toy, x1, y1, x2, y2, r, g, b, _fallback)

    @staticmethod
    def set_led_matrix_line(toy: Toy, x1: int, y1: int, x2: int, y2: int, r: int, g: int, b: int,
                            not_supported_handler: Callable[[], None] = None):
        if toy.implements(IO.draw_compressed_frame_player_line):
            toy.draw_compressed_frame_player_line(x1, y1, x2, y2, r, g, b)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def set_matrix_fill(toy: Toy, x1: int, y1: int, x2: int, y2: int, r: int, g: int, b: int, is_user_color: bool,
                        not_supported_handler: Callable[[], None] = None):
        def _fallback():
            not_supported_handler()

        ToyUtil.set_led_matrix_fill(toy, x1, y1, x2, y2, r, g, b, _fallback)

    @staticmethod
    def set_led_matrix_fill(toy: Toy, x1: int, y1: int, x2: int, y2: int, r: int, g: int, b: int,
                            not_supported_handler: Callable[[], None] = None):
        if toy.implements(IO.draw_compressed_frame_player_fill):
            toy.draw_compressed_frame_player_fill(x1, y1, x2, y2, r, g, b)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def set_matrix_pixel(toy: Toy, x: int, y: int, r: int, g: int, b: int, is_user_color: bool,
                         not_supported_handler: Callable[[], None] = None):
        def _fallback():
            not_supported_handler()

        ToyUtil.set_led_matrix_pixel(toy, x, y, r, g, b, _fallback)

    @staticmethod
    def set_led_matrix_pixel(toy: Toy, x: int, y: int, r: int, g: int, b: int,
                             not_supported_handler: Callable[[], None] = None):
        if toy.implements(IO.set_compressed_frame_player_pixel):
            toy.set_compressed_frame_player_pixel(x, y, r, g, b)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def set_matrix_line(toy: Toy, x1: int, y1: int, x2: int, y2: int, r: int, g: int, b: int, is_user_color: bool,
                        not_supported_handler: Callable[[], None] = None):
        def _fallback():
            not_supported_handler()

        ToyUtil.set_led_matrix_line(toy, x1, y1, x2, y2, r, g, b, _fallback)

    @staticmethod
    def set_led_matrix_line(toy: Toy, x1: int, y1: int, x2: int, y2: int, r: int, g: int, b: int,
                            not_supported_handler: Callable[[], None] = None):
        if toy.implements(IO.draw_compressed_frame_player_line):
            toy.draw_compressed_frame_player_line(x1, y1, x2, y2, r, g, b)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def set_matrix_fill(toy: Toy, x1: int, y1: int, x2: int, y2: int, r: int, g: int, b: int, is_user_color: bool,
                        not_supported_handler: Callable[[], None] = None):
        def _fallback():
            not_supported_handler()

        ToyUtil.set_led_matrix_fill(toy, x1, y1, x2, y2, r, g, b, _fallback)

    @staticmethod
    def set_led_matrix_fill(toy: Toy, x1: int, y1: int, x2: int, y2: int, r: int, g: int, b: int,
                            not_supported_handler: Callable[[], None] = None):
        if toy.implements(IO.draw_compressed_frame_player_fill):
            toy.draw_compressed_frame_player_fill(x1, y1, x2, y2, r, g, b)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def turn_off_leds(toy: Toy):
        if hasattr(toy, 'LEDs'):
            mapping = dict((e, 0) for e in toy.LEDs)
        else:
            mapping = None

        if isinstance(toy, RVR):
            mapping.pop(RVR.LEDs.UNDERCARRIAGE_WHITE)

        def __fallback():
            ToyUtil.set_main_led(toy, 0, 0, 0, False)
            ToyUtil.set_back_led_brightness(toy, 0)

        ToyUtil.set_multiple_leds(toy, mapping, __fallback)
        ToyUtil.set_led_matrix_one_colour(toy, 0, 0, 0)

    # Bolt Animation
    def register_matrix_animation(toy: Toy, s, s2, z, s3, s_arr, i, i_arr,
                                  not_supported_handler: Callable[[], None] = None):
        def _fallback():
            not_supported_handler()
        ToyUtil.save_compressed_frame_player_animation(toy, s, s2, z, s3, s_arr, i, i_arr, _fallback)

    def save_compressed_frame_player_animation(toy: Toy, s, s2, z, s3, s_arr, i, i_arr,
                                               not_supported_handler: Callable[[], None] = None):
        if toy.implements(IO.save_compressed_frame_player_animation):
            toy.save_compressed_frame_player_animation(s, s2, z, s3, s_arr, i, i_arr)
        elif not_supported_handler:
            not_supported_handler()

    def play_matrix_animation(toy: Toy, s, not_supported_handler: Callable[[], None] = None):
        def _fallback():
            not_supported_handler()
        #TODO: Add correct call #ToyUtil.save_compressed_frame_player_animation(toy, s, s2, z, s3, s_arr, i, i_arr, _fallback)

    def play_compressed_frame_player_animation(toy: Toy, s, not_supported_handler: Callable[[], None] = None):
        if toy.implements(IO.play_compressed_frame_player_animation):
            toy.play_compressed_frame_player_animation(s)
        elif not_supported_handler:
            not_supported_handler()

    # Sound
    @staticmethod
    def play_sound(toy: Toy, sound: IntEnum, force_play: bool, not_supported_handler: Callable[[], None] = None):
        if toy.implements(IO.play_audio_file):
            toy.play_audio_file(sound,
                                AudioPlaybackModes.PLAY_IMMEDIATELY if force_play
                                else AudioPlaybackModes.PLAY_ONLY_IF_NOT_PLAYING)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def enable_sensors(toy: Toy, sensors: List[str], not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'sensor_control'):
            toy.sensor_control.enable(*sensors)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def disable_sensors(toy: Toy, not_supported_handler: Callable[[], None] = None):
        if hasattr(toy, 'sensor_control'):
            toy.sensor_control.disable_all()
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def get_ambient_light_sensor_value(toy: Toy):
        return toy.get_ambient_light_sensor_value()

    @staticmethod
    def add_listeners(toy: Toy, manager):
        if hasattr(toy, 'sensor_control') and hasattr(manager, '_sensor_data_listener'):
            toy.sensor_control.add_sensor_data_listener(manager._sensor_data_listener)
        if hasattr(toy, 'add_collision_detected_notify_listener') and hasattr(manager, '_collision_detected_notify'):
            toy.add_collision_detected_notify_listener(manager._collision_detected_notify)
        if hasattr(toy, 'add_battery_state_changed_notify_listener') and \
                hasattr(manager, '_battery_state_changed_notify'):
            toy.add_battery_state_changed_notify_listener(manager._battery_state_changed_notify)
        if hasattr(toy, 'add_gyro_max_notify_listener') and hasattr(manager, '_gyro_max_notify'):
            toy.add_gyro_max_notify_listener(manager._gyro_max_notify)
        if hasattr(toy, 'add_will_sleep_notify_listener') and hasattr(manager, '_will_sleep_notify'):
            toy.add_will_sleep_notify_listener(manager._will_sleep_notify)
        if hasattr(toy, 'add_robot_to_robot_infrared_message_received_notify_listener') and \
                hasattr(manager, '_robot_to_robot_infrared_message_received_notify'):
            toy.add_robot_to_robot_infrared_message_received_notify_listener(
                manager._robot_to_robot_infrared_message_received_notify)
        if hasattr(toy, 'add_sensor_streaming_data_notify_listener'):
            toy.add_sensor_streaming_data_notify_listener(manager._sensor_streaming_data_notify)
        if hasattr(toy, 'add_magnetometer_north_yaw_notify_listener'):
            toy.add_magnetometer_north_yaw_notify_listener(manager._magnetometer_north_yaw_notify)

    @staticmethod
    def set_locator_flags(toy: Toy, flag: bool, not_supported_handler: Callable[[], None] = None):
        if toy.implements(Sensor.set_locator_flags):
            toy.set_locator_flags(flag)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def reset_locator(toy: Toy, not_supported_handler: Callable[[], None] = None):
        if toy.implements(Sphero.configure_locator):
            toy.configure_locator(0, 0, 0, 0)
        elif toy.implements(Sensor.reset_locator_x_and_y):
            toy.reset_locator_x_and_y()
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def configure_collision_detection(toy: Toy, not_supported_handler: Callable[[], None] = None):
        x, y, t = 90, 130, 1
        if toy.implements(Sensor.configure_collision_detection):
            toy.configure_collision_detection(CollisionDetectionMethods.ACCELEROMETER_BASED_DETECTION, x, y, x, y, t)
        elif toy.implements(Sphero.configure_collision_detection):
            toy.configure_collision_detection(SpheroCollisionDetectionMethods.DEFAULT, x, y, x, y, t)
        elif toy.implements(Sensor.configure_sensitivity_based_collision_detection):
            toy.configure_sensitivity_based_collision_detection(
                SensitivityBasedCollisionDetectionMethods.ACCELEROMETER_BASED_DETECTION, SensitivityLevels.VERY_HIGH, t)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def start_robot_to_robot_infrared_broadcasting(toy: Toy, far: int, near: int,
                                                   not_supported_handler: Callable[[], None] = None):
        if toy.implements(Sensor.start_robot_to_robot_infrared_broadcasting):
            toy.start_robot_to_robot_infrared_broadcasting(far, near)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def stop_robot_to_robot_infrared_broadcasting(toy: Toy, not_supported_handler: Callable[[], None] = None):
        if toy.implements(Sensor.stop_robot_to_robot_infrared_broadcasting):
            toy.stop_robot_to_robot_infrared_broadcasting()
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def start_robot_to_robot_infrared_following(toy: Toy, far: int, near: int,
                                                not_supported_handler: Callable[[], None] = None):
        if toy.implements(Sensor.start_robot_to_robot_infrared_following):
            toy.start_robot_to_robot_infrared_following(far, near)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def stop_robot_to_robot_infrared_following(toy: Toy, not_supported_handler: Callable[[], None] = None):
        if toy.implements(Sensor.stop_robot_to_robot_infrared_following):
            toy.stop_robot_to_robot_infrared_following()
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def start_robot_to_robot_infrared_evading(toy: Toy, far: int, near: int,
                                              not_supported_handler: Callable[[], None] = None):
        if toy.implements(Sensor.start_robot_to_robot_infrared_evading):
            toy.start_robot_to_robot_infrared_evading(far, near)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def stop_robot_to_robot_infrared_evading(toy: Toy, not_supported_handler: Callable[[], None] = None):
        if toy.implements(Sensor.stop_robot_to_robot_infrared_evading):
            toy.stop_robot_to_robot_infrared_evading()
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def send_robot_to_robot_infrared_message(toy: Toy, channel: int, intensity: int,
                                             not_supported_handler: Callable[[], None] = None):
        # if toy.implements(Sensor.send_robot_to_robot_infrared_message): TODO: BOLT
        #     toy.send_robot_to_robot_infrared_message(channel, intensity, intensity, intensity, intensity)
        if toy.implements(Sensor.send_infrared_message):
            toy.send_infrared_message(channel, intensity, intensity, intensity, intensity)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def listen_for_robot_to_robot_infrared_message(toy: Toy, channels: Iterable[int], duration: int,
                                                   not_supported_handler: Callable[[], None] = None):
        # TODO: BOLT
        if toy.implements(Sensor.enable_robot_infrared_message_notify):
            toy.enable_robot_infrared_message_notify(True)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def set_power_notifications(toy: Toy, enable: bool, not_supported_handler: Callable[[], None] = None):
        if toy.implements(Power.enable_charger_state_changed_notify):
            toy.enable_charger_state_changed_notify(enable)
        elif toy.implements(Power.enable_battery_state_changed_notify):
            toy.enable_battery_state_changed_notify(enable)
        elif toy.implements(Power.enable_battery_voltage_state_change_notify):
            toy.enable_battery_voltage_state_change_notify(enable)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def set_color_detection(toy: Toy, enable: bool, not_supported_handler: Callable[[], None] = None):
        if toy.implements(Sensor.enable_color_detection):
            toy.enable_color_detection(enable)
        elif not_supported_handler:
            not_supported_handler()

    @staticmethod
    def set_robot_state_on_start(toy: Toy):
        # TODO setUserColour
        ToyUtil.set_head_position(toy, 0)
        ToyUtil.perform_leg_action(toy, R2LegActions.THREE_LEGS)
        ToyUtil.set_locator_flags(toy, False)
        ToyUtil.configure_collision_detection(toy)
        ToyUtil.set_power_notifications(toy, True)
        if toy.implements(Sensor.enable_gyro_max_notify):
            toy.enable_gyro_max_notify(True)
        if hasattr(toy, 'sensor_control'):
            toy.sensor_control.set_interval(150)
        ToyUtil.turn_off_leds(toy)
        if isinstance(toy, RVR):
            ToyUtil.set_color_detection(toy, True)
            ToyUtil.reset_heading(toy)
        ToyUtil.reset_locator(toy)



import struct
from enum import IntEnum, IntFlag

from spherov2.commands import Commands
from spherov2.helper import to_bytes, to_int
from spherov2.listeners.sensor import SensorStreamingMask, CollisionDetected, BotToBotInfraredReadings, \
    RgbcSensorValues, ColorDetection, StreamingServiceData, MotorCurrent, MotorTemperature, \
    MotorThermalProtectionStatus, ThermalProtectionStatus


class CliffDetectionSensorLocations(IntFlag):
    FRONT_LEFT = 0x1  # 0b1
    FRONT_RIGHT = 0x2  # 0b10
    BACK_LEFT = 0x4  # 0b100
    BACK_RIGHT = 0x8  # 0b1000


class CollisionDetectedAxis(IntFlag):
    X_AXIS = 0x1  # 0b1
    Y_AXIS = 0x2  # 0b10


class GyroMaxFlags(IntFlag):
    MAX_PLUS_X = 0x1  # 0b1
    MAX_MINUS_X = 0x2  # 0b10
    MAX_PLUS_Y = 0x4  # 0b100
    MAX_MINUS_Y = 0x8  # 0b1000
    MAX_PLUS_Z = 0x10  # 0b10000
    MAX_MINUS_Z = 0x20  # 0b100000


class InfraredSensorLocations(IntFlag):
    FRONT_LEFT = 0xff  # 0b11111111
    FRONT_RIGHT = 0xff00  # 0b1111111100000000
    BACK_LEFT = 0xff000000  # 0b11111111000000000000000000000000 #How it's displayed in decompiled: -0x1000000
    BACK_RIGHT = 0xff0000  # 0b111111110000000000000000


class LocatorFlags(IntFlag):
    AUTO_CALIBRATE = 0x1  # 0b1


class CollisionDetectionMethods(IntEnum):
    NO_COLLISION_DETECTION = 0
    ACCELEROMETER_BASED_DETECTION = 1
    ACCELEROMETER_BASED_WITH_EXTRA_FILTERING = 2
    HYBRID_ACCELEROMETER_AND_CONTROL_SYSTEM_DETECTION = 3


class MotorIndexes(IntEnum):
    LEFT_MOTOR_INDEX = 0
    RIGHT_MOTOR_INDEX = 1


class SensitivityBasedCollisionDetectionMethods(IntEnum):
    ACCELEROMETER_BASED_DETECTION = 0


class SensitivityLevels(IntEnum):
    SUPER_HIGH = 0
    VERY_HIGH = 1
    HIGH = 2
    MEDIUM = 3
    LOW = 4
    VERY_LOW = 5


class SteamingDataSizes(IntEnum):
    EIGHT_BIT = 0
    SIXTEEN_BIT = 1
    THIRTY_TWO_BIT = 2


class ThermalProtectionStatus(IntEnum):
    OK = 0
    WARN = 1
    CRITICAL = 2


class Sensor(Commands):
    _did = 24

    @staticmethod
    def set_sensor_streaming_mask(toy, interval, count, sensor_masks, proc=None):
        toy._execute(Sensor._encode(toy, 0, proc, [*to_bytes(interval, 2), count, *to_bytes(sensor_masks, 4)]))

    @staticmethod
    def get_sensor_streaming_mask(toy, proc=None):
        return SensorStreamingMask(*struct.unpack('>HBI', toy._execute(Sensor._encode(toy, 1, proc)).data))

    sensor_streaming_data_notify = (
        (24, 2, 0xff), lambda listener, p: listener(list(struct.unpack('>%df' % (len(p.data) // 4), p.data))))

    @staticmethod
    def set_extended_sensor_streaming_mask(toy, sensor_masks, proc=None):
        toy._execute(Sensor._encode(toy, 12, proc, to_bytes(sensor_masks, 4)))

    @staticmethod
    def get_extended_sensor_streaming_mask(toy, proc=None):
        return to_int(toy._execute(Sensor._encode(toy, 13, proc)).data)

    @staticmethod
    def enable_gyro_max_notify(toy, enable, proc=None):
        toy._execute(Sensor._encode(toy, 15, proc, [int(enable)]))

    gyro_max_notify = (24, 16, 0xff), lambda listener, p: listener(p.data[0])

    @staticmethod
    def configure_collision_detection(toy, collision_detection_method: CollisionDetectionMethods,
                                      x_threshold, y_threshold, x_speed, y_speed, dead_time, proc=None):
        toy._execute(Sensor._encode(
            toy, 17, proc, [collision_detection_method, x_threshold, y_threshold, x_speed, y_speed, dead_time]))

    @staticmethod
    def __collision_detected_notify_helper(listener, packet):
        unpacked = struct.unpack('>3hB3hBL', packet.data)
        listener(CollisionDetected(acceleration_x=unpacked[0] / 4096, acceleration_y=unpacked[1] / 4096,
                                   acceleration_z=unpacked[2] / 4096, x_axis=bool(unpacked[3] & 1),
                                   y_axis=bool(unpacked[3] & 2), power_x=unpacked[4], power_y=unpacked[5],
                                   power_z=unpacked[6], speed=unpacked[7], time=unpacked[8] / 1000))

    collision_detected_notify = (24, 18, 0xff), __collision_detected_notify_helper.__func__

    @staticmethod
    def reset_locator_x_and_y(toy, proc=None):
        toy._execute(Sensor._encode(toy, 19, proc))

    @staticmethod
    def enable_collision_detected_notify(toy, enable: bool, proc=None):  # Untested
        toy._execute(Sensor._encode(toy, 20, proc, [int(enable)]))

    @staticmethod
    def set_locator_flags(toy, locator_flags: bool, proc=None):
        toy._execute(Sensor._encode(toy, 23, proc, [int(locator_flags)]))

    @staticmethod
    def set_accelerometer_activity_threshold(toy, threshold: float, proc=None):
        toy._execute(Sensor._encode(toy, 24, proc, struct.pack('>f', threshold)))

    @staticmethod
    def enable_accelerometer_activity_notify(toy, enable: bool, proc=None):
        toy._execute(Sensor._encode(toy, 25, proc, [int(enable)]))

    accelerometer_activity_notify = (24, 26, 0xff), lambda listener, _: listener()

    @staticmethod
    def set_gyro_activity_threshold(toy, threshold: float, proc=None):
        toy._execute(Sensor._encode(toy, 27, proc, struct.pack('>f', threshold)))

    @staticmethod
    def enable_gyro_activity_notify(toy, enable: bool, proc=None):
        toy._execute(Sensor._encode(toy, 28, proc, [int(enable)]))

    gyro_activity_notify = (24, 29, 0xff), lambda listener, _: listener()

    @staticmethod
    def get_bot_to_bot_infrared_readings(toy, proc=None):
        data = to_int(toy._execute(Sensor._encode(toy, 34, proc)).data)
        return BotToBotInfraredReadings(data & 1, data & 2, data & 4, data & 8)

    @staticmethod
    def get_rgbc_sensor_values(toy, proc=None):
        return RgbcSensorValues(*struct.unpack('>4H', toy._execute(Sensor._encode(toy, 35, proc)).data))

    @staticmethod
    def magnetometer_calibrate_to_north(toy, proc=None):
        toy._execute(Sensor.magnetometer_calibrate_to_north(Sensor._encode(toy, 37, proc)))

    magnetometer_north_yaw_notify = (24, 38, 0xff), lambda listener, p: listener(to_int(p.data))

    @staticmethod
    def start_robot_to_robot_infrared_broadcasting(toy, far_code, near_code, proc=None):
        toy._execute(Sensor._encode(toy, 39, proc, [far_code, near_code]))

    @staticmethod
    def start_robot_to_robot_infrared_following(toy, far_code, near_code, proc=None):
        toy._execute(Sensor._encode(toy, 40, proc, [far_code, near_code]))

    @staticmethod
    def stop_robot_to_robot_infrared_broadcasting(toy, proc=None):
        toy._execute(Sensor._encode(toy, 41, proc))

    @staticmethod
    def send_robot_to_robot_infrared_message(toy, s, s2, s3, s4, s5, proc=None):  # Untested / Unknown param names
        toy._execute(Sensor._encode(toy, 42, proc, [s, s2, s3, s4, s5]))

    @staticmethod
    def listen_for_robot_to_robot_infrared_message(toy, s, j, proc=None):  # Untested / Unknown param names
        toy._execute(Sensor._encode(toy, 43, proc, [s, j]))

    robot_to_robot_infrared_message_received_notify = (24, 44, 0xff), lambda listener, p: listener(p.data[0])

    @staticmethod
    def get_ambient_light_sensor_value(toy, proc=None):
        return struct.unpack('>f', toy._execute(Sensor._encode(toy, 48, proc)).data)[0]

    @staticmethod
    def stop_robot_to_robot_infrared_following(toy, proc=None):
        toy._execute(Sensor._encode(toy, 50, proc))

    @staticmethod
    def start_robot_to_robot_infrared_evading(toy, far_code, near_code, proc=None):
        toy._execute(Sensor._encode(toy, 51, proc, [far_code, near_code]))

    @staticmethod
    def stop_robot_to_robot_infrared_evading(toy, proc=None):
        toy._execute(Sensor._encode(toy, 52, proc))

    @staticmethod
    def enable_color_detection_notify(toy, enable, interval, minimum_confidence_threshold, proc=None):
        toy._execute(
            Sensor._encode(toy, 53, proc, [int(enable), *to_bytes(interval, 2), minimum_confidence_threshold]))

    color_detection_notify = (24, 54, 0xff), lambda listener, p: listener(ColorDetection(*p.data))

    @staticmethod
    def get_current_detected_color_reading(toy, proc=None):
        toy._execute(Sensor._encode(toy, 55, proc))

    @staticmethod
    def enable_color_detection(toy, enable, proc=None):
        toy._execute(Sensor._encode(toy, 56, proc, [int(enable)]))

    @staticmethod
    def configure_streaming_service(toy, token, configuration, proc=None):
        toy._execute(Sensor._encode(toy, 57, proc, [token, *configuration]))

    @staticmethod
    def start_streaming_service(toy, period, proc=None):
        toy._execute(Sensor._encode(toy, 58, proc, to_bytes(period, 2)))

    @staticmethod
    def stop_streaming_service(toy, proc=None):
        toy._execute(Sensor._encode(toy, 59, proc))

    @staticmethod
    def clear_streaming_service(toy, proc=None):
        toy._execute(Sensor._encode(toy, 60, proc))

    streaming_service_data_notify = (
        (24, 61, 0xff), lambda listener, p: listener(p.sid, StreamingServiceData(p.data[0], p.data[1:])))

    @staticmethod
    def enable_robot_infrared_message_notify(toy, enable, proc=None):
        toy._execute(Sensor._encode(toy, 62, proc, [int(enable)]))

    @staticmethod
    def send_infrared_message(toy, infrared_code, front_strength, left_strength, right_strength, rear_strength,
                              proc=None):
        toy._execute(Sensor._encode(
            toy, 63, proc, [infrared_code, front_strength, left_strength, right_strength, rear_strength]))

    motor_current_notify = (
        (24, 64, 0xff), lambda listener, p: listener(MotorCurrent(*struct.unpack('>2fQ', p.data))))

    @staticmethod
    def enable_motor_current_notify(toy, enable, proc=None):
        toy._execute(Sensor._encode(toy, 65, proc, data=[int(enable)]))

    @staticmethod
    def get_motor_temperature(toy, motor_index, proc=None):
        return MotorTemperature(
            *struct.unpack('>2f', toy._execute(Sensor._encode(toy, 66, proc, data=[motor_index])).data))

    @staticmethod
    def configure_sensitivity_based_collision_detection(
            toy, method: SensitivityBasedCollisionDetectionMethods, level: SensitivityLevels, i,  # unknown name
            proc=None):
        toy._execute(Sensor._encode(toy, 71, proc, data=[method, level, *to_bytes(i, 2)]))

    @staticmethod
    def enable_sensitivity_based_collision_detection_notify(toy, enable, proc=None):
        toy._execute(Sensor._encode(toy, 72, proc, data=[int(enable)]))

    sensitivity_based_collision_detected_notify = (24, 73, 0xff), lambda listener, p: listener(to_int(p.data))

    @staticmethod
    def get_motor_thermal_protection_status(toy, proc=None):
        data = struct.unpack('>fBfB', toy._execute(Sensor._encode(toy, 75, proc)).data)
        return MotorThermalProtectionStatus(
            data[0], ThermalProtectionStatus(data[1]), data[2], ThermalProtectionStatus(data[3]))

    @staticmethod
    def enable_motor_thermal_protection_status_notify(toy, enable, proc=None):
        toy._execute(Sensor._encode(toy, 76, proc, [int(enable)]))

    @staticmethod
    def __motor_thermal_protection_status_notify_helper(listener, packet):
        data = struct.unpack('>fBfB', packet.data)
        listener(MotorThermalProtectionStatus(
            data[0], ThermalProtectionStatus(data[1]), data[2], ThermalProtectionStatus(data[3])))

    motor_thermal_protection_status_notify = (24, 77, 0xff), __motor_thermal_protection_status_notify_helper.__func__



from enum import IntEnum
from typing import NamedTuple


class MotorIndexes(IntEnum):
    LEFT_MOTOR_INDEX = 0
    RIGHT_MOTOR_INDEX = 1


class MotorStall(NamedTuple):
    motor_index: MotorIndexes
    is_triggered: bool



from spherov2.commands import Commands


class Bootloader(Commands):
    _did = 1

    @staticmethod
    def begin_reflash(toy, proc=None):
        toy._execute(Bootloader._encode(toy, 2, proc))

    @staticmethod
    def here_is_page(toy, proc=None):
        toy._execute(Bootloader._encode(toy, 3, proc))

    @staticmethod
    def jump_to_main(toy, proc=None):
        toy._execute(Bootloader._encode(toy, 4, proc))
